/*************************************************
*    The PMW Music Typesetter - 3rd incarnation  *
*************************************************/

/* Copyright (c) Philip Hazel, 1991 - 2003 */

/* Written by Philip Hazel, starting November 1991 */
/* This file last modified: January 2003 */


/* This file contains code for handling text fonts */


#include "pmwhdr.h"
#include "readhdr.h"


/* We use a built-in table containing the encoding that PMW uses for those 
fonts that normally have the Adobe Standard Encoding. This table must be kept 
in step with the encoding that is defined in the PostScript header file. The 
table is in alphabetic order of name, for fast searching. */

typedef struct encod {
  uschar *name;
  int code;
} encod;

static encod pmwencoding[] = {
  { US"A",                 65 },
  { US"AE",               198 },
  { US"Aacute",           193 },
  { US"Acircumflex",      194 },
  { US"Adieresis",        196 },
  { US"Agrave",           192 },
  { US"Aring",            197 },
  { US"Atilde",           195 },
  { US"B",                 66 },
  { US"C",                 67 },
  { US"Ccedilla",         199 },
  { US"D",                 68 },
  { US"E",                 69 },
  { US"Eacute",           201 },
  { US"Ecircumflex",      202 },
  { US"Edieresis",        203 },
  { US"Egrave",           200 },
  { US"Eth",              208 },
  { US"F",                 70 },
  { US"G",                 71 },
  { US"H",                 72 },
  { US"I",                 73 },
  { US"Iacute",           205 },
  { US"Icircumflex",      206 },
  { US"Idieresis",        207 },
  { US"Igrave",           204 },
  { US"J",                 74 },
  { US"K",                 75 },
  { US"L",                 76 },
  { US"M",                 77 },
  { US"N",                 78 },
  { US"Ntilde",           209 },
  { US"O",                 79 },
  { US"OE",               155 },
  { US"Oacute",           211 },
  { US"Ocircumflex",      212 },
  { US"Odieresis",        214 },
  { US"Ograve",           210 },
  { US"Oslash",           216 },
  { US"Otilde",           213 },
  { US"P",                 80 },
  { US"Q",                 81 },
  { US"R",                 82 },
  { US"S",                 83 },
  { US"T",                 84 },
  { US"Thorn",            222 },
  { US"U",                 85 },
  { US"Uacute",           218 },
  { US"Ucircumflex",      219 },
  { US"Udieresis",        220 },
  { US"Ugrave",           217 },
  { US"V",                 86 },
  { US"W",                 87 },
  { US"X",                 88 },
  { US"Y",                 89 },
  { US"Yacute",           221 },
  { US"Z",                 90 },
  { US"a",                 97 },
  { US"aacute",           225 },
  { US"acircumflex",      226 },
  { US"acute",            180 },
  { US"adieresis",        228 },
  { US"ae",               230 },
  { US"agrave",           224 },
  { US"ampersand",         38 },
  { US"aring",            229 },
  { US"asciicircum",       94 },
  { US"asciitilde",       126 },
  { US"asterisk",          42 },
  { US"at",                64 },
  { US"atilde",           227 },
  { US"b",                 98 },
  { US"backslash",         92 },
  { US"bar",              124 },
  { US"braceleft",        123 },
  { US"braceright",       125 },
  { US"bracketleft",       91 },
  { US"bracketright",      93 },
  { US"brokenbar",        166 },
  { US"bullet",           143 },
  { US"c",                 99 },
  { US"ccedilla",         231 },
  { US"cedilla",          184 },
  { US"cent",             162 },
  { US"circumflex",        29 },
  { US"colon",             58 },
  { US"comma",             44 },
  { US"copyright",        169 },
  { US"currency",         164 },
  { US"d",                100 },
  { US"dagger",           157 },
  { US"degree",           176 },
  { US"dieresis",         168 },
  { US"divide",           247 },
  { US"dollar",            36 },
  { US"dotlessi",          31 },
  { US"e",                101 },
  { US"eacute",           233 },
  { US"ecircumflex",      234 },
  { US"edieresis",        235 },
  { US"egrave",           232 },
  { US"eight",             56 },
  { US"emdash",           153 },
  { US"endash",           152 },
  { US"equal",             61 },
  { US"eth",              240 },
  { US"exclam",            33 },
  { US"exclamdown",       161 },
  { US"f",                102 },
  { US"fi",               158 },
  { US"five",              53 },
  { US"fl",               159 },
  { US"four",              52 },
  { US"g",                103 },
  { US"germandbls",       223 },
  { US"grave",             96 },
  { US"greater",           62 },
  { US"guillemotleft",    171 },
  { US"guillemotright",   187 },
  { US"guilsinglleft",    146 },
  { US"guilsinglright",   147 },
  { US"h",                104 },
  { US"hardspace",        160 },
  { US"hyphen",            45 },
  { US"i",                105 },
  { US"iacute",           237 },
  { US"icircumflex",      238 },
  { US"idieresis",        239 },
  { US"igrave",           236 },
  { US"j",                106 },
  { US"k",                107 },
  { US"l",                108 },
  { US"less",              60 },
  { US"logicalnot",       172 },
  { US"m",                109 },
  { US"macron",           175 },
  { US"minus",            154 },
  { US"mu",               181 },
  { US"multiply",         215 },
  { US"n",                110 },
  { US"nine",              57 },
  { US"ntilde",           241 },
  { US"numbersign",        35 },
  { US"o",                111 },
  { US"oacute",           243 },
  { US"ocircumflex",      244 },
  { US"odieresis",        246 },
  { US"oe",               156 },
  { US"ograve",           242 },
  { US"one",               49 },
  { US"onehalf",          189 },
  { US"onequarter",       188 },
  { US"onesuperior",      185 },
  { US"ordfeminine",      170 },
  { US"ordmasculine",     186 },
  { US"oslash",           248 },
  { US"otilde",           245 },
  { US"p",                112 },
  { US"paragraph",        182 },
  { US"parenleft",         40 },
  { US"parenright",        41 },
  { US"percent",           37 },
  { US"period",            46 },
  { US"periodcentered",   183 },
  { US"plus",              43 },
  { US"plusminus",        177 },
  { US"q",                113 },
  { US"question",          63 },
  { US"questiondown",     191 },
  { US"quotedbl",          34 },
  { US"quotedblbase",     150 },
  { US"quotedblleft",     148 },
  { US"quotedblright",    149 },
  { US"quoteleft",        144 },
  { US"quoteright",       145 },
  { US"quotesingle",       39 },
  { US"r",                114 },
  { US"registered",       174 },
  { US"s",                115 },
  { US"section",          167 },
  { US"semicolon",         59 },
  { US"seven",             55 },
  { US"six",               54 },
  { US"slash",             47 },
  { US"space",             32 },
  { US"sterling",         163 },
  { US"t",                116 },
  { US"thorn",            254 },
  { US"three",             51 },
  { US"threequarters",    190 },
  { US"threesuperior",    179 },
  { US"tilde",             28 },
  { US"two",               50 },
  { US"twosuperior",      178 },
  { US"u",                117 },
  { US"uacute",           250 },
  { US"ucircumflex",      251 },
  { US"udieresis",        252 },
  { US"ugrave",           249 },
  { US"underscore",        95 },
  { US"v",                118 },
  { US"w",                119 },
  { US"x",                120 },
  { US"y",                121 },
  { US"yacute",           253 },
  { US"ydieresis",        255 },
  { US"yen",              165 },
  { US"z",                122 },
  { US"zero",              48 }};

static int pmwencodingcount = sizeof(pmwencoding)/sizeof(encod);



/*************************************************
*        Kern table sorting comparison           *
*************************************************/

/* This is the auxiliary routine used for comparing kern
table entries when sorting them. */

static int table_cmp(const void *a, const void *b)
{
return *((int *)a) - *((int *)b);
}



/*************************************************
*          Number reader from AFM files          *
*************************************************/

static uschar *read_number(int *value, uschar *p)
{
int n = 0;
int sign = 1;
while (*p != 0 && *p == ' ') p++;
if (*p == '-') { sign = -1; p++; }
while (isdigit(*p)) n = n * 10 + *p++ - '0';
*value = n * sign;
return p;
}


/*************************************************
*      Load width and kern tables for a font     *
*************************************************/

static void font_loadtables(int fontid)
{
FILE *f;
BOOL usepmwencoding = FALSE;
int *widths, *heights;
fontstr *fs = &(font_List[fontid]);
uschar *pp;
uschar filename[256];
uschar line[256];

sprintf(CS filename, "%s/%s.afm", font_metrics, fs->psname);
f = Ufopen(filename, "r");
if (f == NULL) error_moan(59, fs->psname, font_metrics);   /* Hard */

widths = fs->widths = malloc(256 * sizeof(int));
memset(widths, 0, 256 * sizeof(int));
heights = fs->heights = NULL;
fs->kerncount = 0;

/* Process the AFM file. First find the start of the metrics; on the way, check 
for the standard encoding scheme. For those fonts, we use the PMW encoding 
scheme. */

for (;;)
  {
  if (Ufgets(line, sizeof(line), f) == NULL)
    error_moan(122, filename, "no metric data found", "");  /* Hard */
  if (memcmp(line, "EncodingScheme AdobeStandardEncoding", 36) == 0)
    usepmwencoding = TRUE;
  if (memcmp(line, "StartCharMetrics", 16) == 0) break;
  }

/* Process the metric lines for each character */

for (;;)
  {
  int width;
  uschar *cname;
 
  if (Ufgets(line, sizeof(line), f) == NULL)
    error_moan(122, filename, "unexpected end of metric data", "");
  if (memcmp(line, "EndCharMetrics", 14) == 0) break;

  if (memcmp(line, "C ", 2) != 0)
    error_moan(122, filename, "unrecognized metric data line: ", line);
  
  pp = line + 2;
  while (memcmp(pp, "WX", 2) != 0) pp++;
  pp = read_number(&width, pp+2);

  /* If the is a font for which we are going to use PMW encoding (i.e. it is a 
  StandardEncoding font), scan encoding file for this character. If it is not
  present, just ignore it. */
  
  if (usepmwencoding)
    {  
    int top, bot; 
    while (memcmp(pp, "N ", 2) != 0) pp++;
    cname = (pp += 2);
    while (*pp != ' ') pp++;
    *pp = 0;
    
    bot = 0;
    top = pmwencodingcount;
    while (top > bot)
      {
      int mid = (top + bot)/2;
      int c = Ustrcmp(cname, pmwencoding[mid].name);
      if (c == 0)
        {
        widths[pmwencoding[mid].code] = width;
        break;
        }
      if (c > 0) bot = mid + 1; else top = mid;       
      }    
    }
 
  /* For other fonts, just use the character number directly. These fonts 
  include the PMW-Music font, which has some characters with vertical height 
  movements. */
  
  else
    {
    int n;
    (void)read_number(&n, line+1);
    widths[n] = width;  
    while (*pp != 0 && memcmp(pp, "WY", 2) != 0) pp++;
    if (*pp != 0)
      {
      int height; 
      pp += 2;
      (void)read_number(&height, pp);
      if (heights == NULL)
        {
        heights = fs->heights = malloc(256 * sizeof(int)); 
        memset(heights, 0, 256 * sizeof(int));
        }  
      heights[n] = height; 
      } 
    }  
  }

/* If the font had a standard encoding and we are using kerning, process
kerning data (if any); when this is done, we are finished with the AFM file, so
can stop. */

if (usepmwencoding && main_kerning)
  {
  int finalcount = 0;
  int kerncount = 0;
  int *kerntable;

  for (;;)
    {
    if (Ufgets(line, sizeof(line), f) == NULL)
      {
      fclose(f);
      return;
      }    
    if (memcmp(line, "StartKernPairs", 14) == 0) break;
    }
    
  /* Find size of kern table, and get space for it. Adobe's AFM files
  seem to have a habit of containing a large number of kern pairs with
  zero amount of kern. We leave these out of the table and adjust
  the count for searching, but don't bother to free up the unused
  store (it isn't a vast amount). */
  
  pp = line + 14;
  while (*pp != 0 && *pp == ' ') pp++;
  (void) read_number(&kerncount, pp);
  fs->kerns = kerntable = malloc(2*kerncount*sizeof(int));
    
  while (kerncount--)
    {
    int sign = 1; 
    int i, value; 
    int a = -1;
    int b = -1;  
     
    if (Ufgets(line, sizeof(line), f) == NULL)
      error_moan(122, filename, "unexpected end of kerning data", "");
    if (memcmp(line, "EndKernPairs", 12) == 0) break;
    
    /* Skip blank lines */
    
    if (Ustrlen(line) <= 1)
      {
      kerncount++;
      continue;
      }     
    
    /* Process each kern, using the encoding table to determine
    the character numbers. It is assumed that the right encoding
    table has previously been read in. */
    
    pp = line + 4;
    
    for (i = 0; i < 2; i++)
      { 
      int bot = 0;
      int top = pmwencodingcount;
      uschar *x = pp; 
      while (*pp != 0 && *pp != ' ') pp++;
      *pp++ = 0;
      while (*pp != 0 && *pp == ' ') pp++;
    
      /* Now scan the encoding vector to get the character numbers,
      and create the kern index value. */
      
      while (top > bot)
        {
        int mid = (top + bot)/2;
        int c = Ustrcmp(x, pmwencoding[mid].name);
        if (c == 0)
          {
          if (i == 0) a = pmwencoding[mid].code; 
            else b = pmwencoding[mid].code;
          break;
          }
        if (c > 0) bot = mid + 1; else top = mid;       
        }    
      }
      
    /* If we haven't found the characters, ignore the kern */
    
    if (a >= 0 && b >= 0)
      {   
      kerntable[finalcount] = (a << 8) + b;
    
      /* Read the kern value */
      
      while (*pp != 0 && *pp == ' ') pp++;
      if (*pp == '-') { sign = -1; pp++; }
      (void)read_number(&value, pp);
      if (value != 0)
        {
        kerntable[finalcount+1] = value*sign;
        finalcount += 2;
        }
      }
    } 
    
  /* Adjust the count and sort into ascending order */
    
  fs->kerncount = finalcount / 2;  /* true count */
  qsort(kerntable, fs->kerncount, 2*sizeof(int), table_cmp);
  }    

fclose(f);


/* Early checking debugging */
#ifdef NEVER
debug_printf("FONT id=%d %s\n", fontid, fs->psname);
  {
  int i;
  for (i = 0; i < 256; i++)
    {
    debug_printf("%3d %5d", i, fs->widths[i]);
    if (fs->heights != NULL) debug_printf(" %5d", fs->heights[i]);
    debug_printf("\n");  
    }
  debug_printf("KERNS %d %d\n", fs->kerncount, fs->kerns);
  for (i = 0; i < fs->kerncount; i++)
    {
    int c = fs->kerns[2*i];
    int v = fs->kerns[2*i + 1];
    debug_printf("%3d %x %3d (%c) %3d (%c) %d\n", i, c,
      (c >> 8) & 255, (c >> 8) & 255, 
      c & 255, c & 255, v);   
    }   
  }  
#endif
}




/*************************************************
*         Add a font to the base list            *
*************************************************/

void font_addbasefont(uschar *psfont)
{
fontstr *fs = &(font_List[font_basecount]);
if (font_basecount >= max_fonts) { error_moan(26, max_fonts); return; }
font_basecount++;
fs->psname = malloc(Ustrlen(psfont)+1);
Ustrcpy(fs->psname, psfont);
fs->widths = NULL;
fs->heights = NULL;
fs->kerns = NULL;
fs->kerncount = -1;
}


/*************************************************
*        Read the list of text fonts             *
*************************************************/


void font_init(void)
{
font_addbasefont(US"Times-Roman");
font_addbasefont(US"Times-Italic");
font_addbasefont(US"Times-Bold");
font_addbasefont(US"Times-BoldItalic");
font_addbasefont(US"Symbol");
font_addbasefont(US"PMW-Music");
}



/*************************************************
*              Search for given font             *
*************************************************/


int font_search(uschar *name)
{
int i;
for (i = 0; i < font_count; i++)
  if (Ustrcmp(name, font_List[i].psname) == 0) return i;
return -1;
}



/*************************************************
*         Find the width of a string             *
*************************************************/

/* The "height" (distance current point moves vertically) is placed in 
font_stringheight. This is only non-zero when there's a rotation. If 
font_xstretch is set, it is an absolute amount of space to add to each space 
character. This is used when printing justified headings and footings. */

int font_stringwidth(uschar *s, int font, int pointsize)
{
int hwidth = 0;
int vheight = 0;
int spacecount = 0;
int *widths, *heights, *ktable;
int fontid = font_table[font];
if (font_List[fontid].widths == NULL) font_loadtables(fontid);
widths = font_List[fontid].widths;
heights = font_List[fontid].heights;
ktable = font_List[fontid].kerns;

DEBUG(("font_stringwidth %d %d \"%s\"\n", fontid, pointsize, s));

/* When kerning we have to note the kern values. */

if (main_kerning && ktable != NULL)
  {
  int c; 
  int kerncount = font_List[fontid].kerncount; 
  while ((c = *s++) != 0)
    {
    if (c == ' ') spacecount++; 
    hwidth += widths[c];
    if (heights != NULL) vheight += heights[c]; 
    if (*s != 0)
      {
      int bot = 0;
      int top = kerncount;
      c = (c << 8) + *s;
      while (bot < top)
        {
        int mid = (bot + top)/2;
        int d = ktable[mid*2]; 
        if (c == d)
          {
          hwidth += ktable[mid*2+1];
          break;
          }    
        if (c < d) top = mid; else bot = mid + 1;
        }
      }  
    }  
  }
  
/* Not kerning; just add up the widths */
 
else while (*s != 0) 
  {
  int c = *s++; 
  if (c == ' ') spacecount++; 
  if (heights != NULL) vheight += heights[c]; 
  hwidth += widths[c];
  } 

/* Scale to font size, then adjust for rotation */
  
hwidth = mac_muldiv(hwidth, pointsize, 1000);
vheight = mac_muldiv(vheight, pointsize, 1000);

font_stringheight = 
  mac_muldiv(hwidth, font_transform[1], 65536) +
  mac_muldiv(vheight, font_transform[3], 65536);
  
DEBUG(("font_stringwidth() %d %d\n",  
       mac_muldiv(hwidth, font_transform[0], 65536) +
       mac_muldiv(vheight, font_transform[2], 65536),
       font_stringheight)); 

return font_xstretch * spacecount +
       mac_muldiv(hwidth, font_transform[0], 65536) +
       mac_muldiv(vheight, font_transform[2], 65536);
}




/*************************************************
*            Read a font word (+ number)         *
*************************************************/

int font_fontword(BOOL soft)
{
uschar word[100];
read_word(word);
if (strcmp(word, "roman") == 0) return font_rm;
if (strcmp(word, "italic") == 0) return font_it;
if (strcmp(word, "bold") == 0) return font_bf;
if (strcmp(word, "bolditalic") == 0) return font_bi;
if (strcmp(word, "symbol") == 0) return font_sy;
if (strcmp(word, "music") == 0) return font_mu;
if (strcmp(word, "bigmusic") == 0) return font_mf;
if (strcmp(word, "extra") == 0)
  {
  int x;
  if (!read_expect_integer(&x, FALSE, FALSE)) return -1;
  if (x >= 1 && x <= MaxExtraFont) return font_xx + x - 1;
  error_moan(26, MaxExtraFont);
  }
else if (!soft)
  error_moan(10, "\"roman\", \"italic\", \"bold\", \"bolditalic\", or \"extra\"");
return -1;
}



/*************************************************
*         Set up a font rotation                 *
*************************************************/

void font_rotate(int rotate)
{
int newmatrix[4];
double sr = sin(((double)rotate)*atan(1.0)/45000.0);
double cr = cos(((double)rotate)*atan(1.0)/45000.0);

font_sinr = (int)(sr * 1000.0);
font_cosr = (int)(cr * 1000.0);

newmatrix[0] =
  (int)(((double)font_transform[0])*cr - ((double)font_transform[1])*sr);
newmatrix[1] =
  (int)(((double)font_transform[0])*sr + ((double)font_transform[1])*cr);
newmatrix[2] =
  (int)(((double)font_transform[2])*cr - ((double)font_transform[3])*sr);
newmatrix[3] =
  (int)(((double)font_transform[2])*sr + ((double)font_transform[3])*cr);

memcpy(font_transform, newmatrix, 4*sizeof(int));
}



/*************************************************
*           Reset font transformation            *
*************************************************/

/* We operate here in the units that RISC OS uses for its
font transformations so that the screen can use the matrix
directly. The first four values have 16-bit fractions; the
other two are in millipoints. */

void font_reset(void)
{
font_sinr = 0;
font_cosr = 1000;
font_transform[0] = font_transform[3] = 65536;
font_transform[1] = font_transform[2] = 0;
font_transform[4] = font_transform[5] = 0;
}

/* End of font.c */
