/*************************************************
*    The PMW Music Typesetter - 3rd incarnation  *
*************************************************/

/* Copyright (c) Philip Hazel, 1991 - 2002 */

/* Written by Philip Hazel, starting November 1991 */
/* This file last modified: October 2003 */


/* This file contains initializing and other housekeeping functions. */


#include "pmwhdr.h"
#include "outhdr.h"



/*************************************************
*         Read a MIDI translation file           *
*************************************************/

/* These files are short: reading them twice in order to get the
correct size doesn't take much time and saves much hassle. */

void read_midi_translation(uschar **anchor, uschar *filename)
{
FILE *f = Ufopen(filename, "r");
int length = 0;
uschar *p;
uschar line[60];

if (f == NULL) return;

while (Ufgets(line, 60, f) != NULL)
  {
  line[Ustrlen(line)-1] = 0;
  if (!isdigit(line[0])) continue;
  length += Ustrlen(line+4) + 2;
  }

if (length == 0) return;
*anchor = malloc(length+1);
p = *anchor;

rewind(f);
while (Ufgets(line, 60, f) != NULL)
  {
  line[Ustrlen(line)-1] = 0;
  if (!isdigit(line[0])) continue;
  Ustrcpy(p, line+4);
  p += Ustrlen(p) + 1;
  *p++ = Uatoi(line);   
  }
*p = 0;
fclose(f);
}


/*************************************************
*              Local initialization              *
*************************************************/

/* This is called very early, to allow for setting up of any store-handling
things, and anything else that must be done before other initialization
takes place. */

void sys_init1(void)
{
font_metrics = US FONTMETRICS;
ps_header = US PSHEADER;
midi_voices = US MIDIVOICES;
midi_perc = US MIDIPERC;
}



/* This is called before argument decoding is done. */

void sys_init2(void)
{
}



/*************************************************
*            Relativize a file name              *
*************************************************/

/* The name must be in a buffer that is long enough to take the additional path 
if necessary. */

void sys_relativize(uschar *name)
{
uschar temp[256];
int i;
DEBUG(("sys_relativize(%s) entered\n", name));
if (name[0] == '/' || main_filename == NULL) return;
i = Ustrlen(main_filename);
Ustrcpy(temp, name);
while (i > 0 && main_filename[--i] != '/');
if (i != 0) i++;
Ustrncpy(name, main_filename, i);
Ustrcpy(name+i, temp);
DEBUG(("relativized to %s\n", name));
}



/*************************************************
*            Deal with hourglass                 *
*************************************************/

void sys_visdelay(int flag)
{

}


/* End of init.c */
